@@ -28,12 +28,11 @@ we are still 40-60% behind Skia built for Haswell.
 On ARM AArch64 the story is different and explicit SIMD make our code up to 2-3x faster.
 */
 
-use tiny_skia_path::ScreenIntRect;
-
 use crate::PremultipliedColorU8;
 
 use crate::pixmap::SubPixmapMut;
 use crate::wide::{f32x8, u16x16, f32x16};
+use crate::geom::ScreenIntRect;
 
 pub const STAGE_WIDTH: usize = 16;
 
@@ -43,8 +42,8 @@ pub struct Pipeline<'a, 'b: 'a> {
     index: usize,
     functions: &'a [StageFn],
     pixmap: &'a mut SubPixmapMut<'b>,
-    clip_mask_ctx: super::ClipMaskCtx<'a>,
-    mask_ctx: super::AAMaskCtx,
+    mask_ctx: super::MaskCtx<'a>,
+    aa_mask_ctx: super::AAMaskCtx,
     ctx: &'a mut super::Context,
     r: u16x16,
     g: u16x16,
@@ -80,7 +79,10 @@ pub const STAGES: &[StageFn; super::STAGES_COUNT] = &[
     seed_shader,
     load_dst,
     store,
+    load_dst_u8,
+    store_u8,
     null_fn, // Gather
+    load_mask_u8,
     mask_u8,
     scale_u8,
     lerp_u8,
@@ -145,8 +147,8 @@ pub fn start(
     functions: &[StageFn],
     functions_tail: &[StageFn],
     rect: &ScreenIntRect,
-    mask_ctx: super::AAMaskCtx,
-    clip_mask_ctx: super::ClipMaskCtx,
+    aa_mask_ctx: super::AAMaskCtx,
+    mask_ctx: super::MaskCtx,
     ctx: &mut super::Context,
     pixmap: &mut SubPixmapMut,
 ) {
@@ -154,8 +156,8 @@ pub fn start(
         index: 0,
         functions: &[],
         pixmap,
-        clip_mask_ctx,
         mask_ctx,
+        aa_mask_ctx,
         ctx,
         r: u16x16::default(),
         g: u16x16::default(),
@@ -265,12 +267,87 @@ pub fn store_tail(p: &mut Pipeline) {
     p.next_stage();
 }
 
-fn mask_u8(p: &mut Pipeline) {
-    let offset = p.clip_mask_ctx.offset(p.dx, p.dy);
+pub fn load_dst_u8(p: &mut Pipeline) {
+    load_8(p.pixmap.slice16_mask_at_xy(p.dx, p.dy), &mut p.da);
+    p.next_stage();
+}
+
+pub fn load_dst_u8_tail(p: &mut Pipeline) {
+    // Fill a dummy array with `tail` values. `tail` is always in a 1..STAGE_WIDTH-1 range.
+    // This way we can reuse the `load_8888__` method and remove any branches.
+    let data = p.pixmap.slice_mask_at_xy(p.dx, p.dy);
+    let mut tmp = [0u8; STAGE_WIDTH];
+    tmp[0..p.tail].copy_from_slice(&data[0..p.tail]);
+    load_8(&tmp, &mut p.da);
+
+    p.next_stage();
+}
+
+pub fn store_u8(p: &mut Pipeline) {
+    let data = p.pixmap.slice16_mask_at_xy(p.dx, p.dy);
+    let a = p.a.as_slice();
+
+    data[ 0] = a[ 0] as u8;
+    data[ 1] = a[ 1] as u8;
+    data[ 2] = a[ 2] as u8;
+    data[ 3] = a[ 3] as u8;
+    data[ 4] = a[ 4] as u8;
+    data[ 5] = a[ 5] as u8;
+    data[ 6] = a[ 6] as u8;
+    data[ 7] = a[ 7] as u8;
+    data[ 8] = a[ 8] as u8;
+    data[ 9] = a[ 9] as u8;
+    data[10] = a[10] as u8;
+    data[11] = a[11] as u8;
+    data[12] = a[12] as u8;
+    data[13] = a[13] as u8;
+    data[14] = a[14] as u8;
+    data[15] = a[15] as u8;
+
+    p.next_stage();
+}
+
+pub fn store_u8_tail(p: &mut Pipeline) {
+    let data = p.pixmap.slice_mask_at_xy(p.dx, p.dy);
+    let a = p.a.as_slice();
+
+    // This is better than `for i in 0..tail`, because this way the compiler
+    // knows that we have only 16 steps and slices access is guarantee to be valid.
+    // This removes bounds checking and a possible panic call.
+    for i in 0..STAGE_WIDTH {
+        data[i] = a[i] as u8;
+
+        if i + 1 == p.tail {
+            break;
+        }
+    }
+
+    p.next_stage();
+}
+
+// Similar to mask_u8, but only loads the mask values without actually masking the pipeline.
+fn load_mask_u8(p: &mut Pipeline) {
+    let offset = p.mask_ctx.offset(p.dx, p.dy);
 
     let mut c = u16x16::default();
     for i in 0..p.tail {
-        c.0[i] = u16::from(p.clip_mask_ctx.data[offset + i]);
+        c.0[i] = u16::from(p.mask_ctx.data[offset + i]);
+    }
+
+    p.r = u16x16::splat(0);
+    p.g = u16x16::splat(0);
+    p.b = u16x16::splat(0);
+    p.a = c;
+
+    p.next_stage();
+}
+
+fn mask_u8(p: &mut Pipeline) {
+    let offset = p.mask_ctx.offset(p.dx, p.dy);
+
+    let mut c = u16x16::default();
+    for i in 0..p.tail {
+        c.0[i] = u16::from(p.mask_ctx.data[offset + i]);
     }
 
     if c == u16x16::default() {
@@ -287,7 +364,7 @@ fn mask_u8(p: &mut Pipeline) {
 
 fn scale_u8(p: &mut Pipeline) {
     // Load u8xTail and cast it to u16x16.
-    let data = p.mask_ctx.copy_at_xy(p.dx, p.dy, p.tail);
+    let data = p.aa_mask_ctx.copy_at_xy(p.dx, p.dy, p.tail);
     let c = u16x16([
         u16::from(data[0]),
         u16::from(data[1]),
@@ -317,7 +394,7 @@ fn scale_u8(p: &mut Pipeline) {
 
 fn lerp_u8(p: &mut Pipeline) {
     // Load u8xTail and cast it to u16x16.
-    let data = p.mask_ctx.copy_at_xy(p.dx, p.dy, p.tail);
+    let data = p.aa_mask_ctx.copy_at_xy(p.dx, p.dy, p.tail);
     let c = u16x16([
         u16::from(data[0]),
         u16::from(data[1]),
@@ -739,11 +816,22 @@ fn store_8888_tail(
     }
 }
 
+#[inline(always)]
+fn load_8(data: &[u8; STAGE_WIDTH], a: &mut u16x16) {
+    *a = u16x16([
+        data[ 0] as u16, data[ 1] as u16, data[ 2] as u16, data[ 3] as u16,
+        data[ 4] as u16, data[ 5] as u16, data[ 6] as u16, data[ 7] as u16,
+        data[ 8] as u16, data[ 9] as u16, data[10] as u16, data[11] as u16,
+        data[12] as u16, data[13] as u16, data[14] as u16, data[15] as u16,
+    ]);
+}
+
 #[inline(always)]
 fn div255(v: u16x16) -> u16x16 {
     // Skia uses `vrshrq_n_u16(vrsraq_n_u16(v, v, 8), 8)` here when NEON is available,
     // but it doesn't affect performance much and breaks reproducible result. Ignore it.
-    (v + u16x16::splat(255)) / u16x16::splat(256)
+    // NOTE: the compiler does not replace the devision with a shift.
+    (v + u16x16::splat(255)) >> u16x16::splat(8) // / u16x16::splat(256)
 }
 
 #[inline(always)]
